@@ -1,6 +1,7 @@
 package consul
 
 import (
+	"context"
 	"crypto/tls"
 	"errors"
 	"fmt"
@@ -17,7 +18,7 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/hashicorp/consul/acl"
+	ca "github.com/hashicorp/consul/agent/connect/ca"
 	"github.com/hashicorp/consul/agent/consul/autopilot"
 	"github.com/hashicorp/consul/agent/consul/fsm"
 	"github.com/hashicorp/consul/agent/consul/state"
@@ -72,6 +73,15 @@ const (
 	// raftRemoveGracePeriod is how long we wait to allow a RemovePeer
 	// to replicate to gracefully leave the cluster.
 	raftRemoveGracePeriod = 5 * time.Second
+
+	// serfEventChSize is the size of the buffered channel to get Serf
+	// events. If this is exhausted we will block Serf and Memberlist.
+	serfEventChSize = 2048
+
+	// reconcileChSize is the size of the buffered channel reconcile updates
+	// from Serf with the Catalog. If this is exhausted we will drop updates,
+	// and wait for a periodic reconcile.
+	reconcileChSize = 256
 )
 
 var (
@@ -84,11 +94,24 @@ type Server struct {
 	// sentinel is the Sentinel code engine (can be nil).
 	sentinel sentinel.Evaluator
 
-	// aclAuthCache is the authoritative ACL cache.
-	aclAuthCache *acl.Cache
+	// acls is used to resolve tokens to effective policies
+	acls *ACLResolver
 
-	// aclCache is the non-authoritative ACL cache.
-	aclCache *aclCache
+	// aclUpgradeCancel is used to cancel the ACL upgrade goroutine when we
+	// lose leadership
+	aclUpgradeCancel  context.CancelFunc
+	aclUpgradeLock    sync.RWMutex
+	aclUpgradeEnabled bool
+
+	// aclReplicationCancel is used to shut down the ACL replication goroutine
+	// when we lose leadership
+	aclReplicationCancel  context.CancelFunc
+	aclReplicationLock    sync.RWMutex
+	aclReplicationEnabled bool
+
+	// DEPRECATED (ACL-Legacy-Compat) - only needed while we support both
+	// useNewACLs is used to determine whether we can use new ACLs or not
+	useNewACLs int32
 
 	// autopilot is the Autopilot instance for this server.
 	autopilot *autopilot.Autopilot
@@ -96,6 +119,22 @@ type Server struct {
 	// autopilotWaitGroup is used to block until Autopilot shuts down.
 	autopilotWaitGroup sync.WaitGroup
 
+	// caProvider is the current CA provider in use for Connect. This is
+	// only non-nil when we are the leader.
+	caProvider ca.Provider
+	// caProviderRoot is the CARoot that was stored along with the ca.Provider
+	// active. It's only updated in lock-step with the caProvider. This prevents
+	// races between state updates to active roots and the fetch of the provider
+	// instance.
+	caProviderRoot *structs.CARoot
+	caProviderLock sync.RWMutex
+
+	// caPruningCh is used to shut down the CA root pruning goroutine when we
+	// lose leadership.
+	caPruningCh      chan struct{}
+	caPruningLock    sync.RWMutex
+	caPruningEnabled bool
+
 	// Consul configuration
 	config *Config
 
@@ -208,15 +247,24 @@ type Server struct {
 	shutdown     bool
 	shutdownCh   chan struct{}
 	shutdownLock sync.Mutex
+
+	// embedded struct to hold all the enterprise specific data
+	EnterpriseServer
 }
 
+// NewServer is only used to help setting up a server for testing. Normal code
+// exercises NewServerLogger.
 func NewServer(config *Config) (*Server, error) {
-	return NewServerLogger(config, nil, new(token.Store))
+	c, err := tlsutil.NewConfigurator(config.ToTLSUtilConfig(), nil)
+	if err != nil {
+		return nil, err
+	}
+	return NewServerLogger(config, nil, new(token.Store), c)
 }
 
-// NewServer is used to construct a new Consul server from the
+// NewServerLogger is used to construct a new Consul server from the
 // configuration, potentially returning an error
-func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*Server, error) {
+func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store, tlsConfigurator *tlsutil.Configurator) (*Server, error) {
 	// Check the protocol version.
 	if err := config.CheckProtocolVersion(); err != nil {
 		return nil, err
@@ -245,17 +293,13 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		config.UseTLS = true
 	}
 
-	// Create the TLS wrapper for outgoing connections.
-	tlsConf := config.tlsConfig()
-	tlsWrap, err := tlsConf.OutgoingTLSWrapper()
-	if err != nil {
-		return nil, err
-	}
-
-	// Get the incoming TLS config.
-	incomingTLS, err := tlsConf.IncomingTLSConfig()
-	if err != nil {
-		return nil, err
+	// Set the primary DC if it wasn't set.
+	if config.PrimaryDatacenter == "" {
+		if config.ACLDatacenter != "" {
+			config.PrimaryDatacenter = config.ACLDatacenter
+		} else {
+			config.PrimaryDatacenter = config.Datacenter
+		}
 	}
 
 	// Create the tombstone GC.
@@ -272,7 +316,7 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		LogOutput:  config.LogOutput,
 		MaxTime:    serverRPCCache,
 		MaxStreams: serverMaxStreams,
-		TLSWrapper: tlsWrap,
+		TLSWrapper: tlsConfigurator.OutgoingRPCWrapper(),
 		ForceTLS:   config.VerifyOutgoing,
 	}
 
@@ -281,14 +325,14 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		config:           config,
 		tokens:           tokens,
 		connPool:         connPool,
-		eventChLAN:       make(chan serf.Event, 256),
-		eventChWAN:       make(chan serf.Event, 256),
+		eventChLAN:       make(chan serf.Event, serfEventChSize),
+		eventChWAN:       make(chan serf.Event, serfEventChSize),
 		logger:           logger,
 		leaveCh:          make(chan struct{}),
-		reconcileCh:      make(chan serf.Member, 32),
+		reconcileCh:      make(chan serf.Member, reconcileChSize),
 		router:           router.NewRouter(logger, config.Datacenter),
 		rpcServer:        rpc.NewServer(),
-		rpcTLS:           incomingTLS,
+		rpcTLS:           tlsConfigurator.IncomingRPCConfig(),
 		reassertLeaderCh: make(chan chan error),
 		segmentLAN:       make(map[string]*serf.Serf, len(config.Segments)),
 		sessionTimers:    NewSessionTimers(),
@@ -297,30 +341,33 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		shutdownCh:       shutdownCh,
 	}
 
+	// Initialize enterprise specific server functionality
+	if err := s.initEnterprise(); err != nil {
+		s.Shutdown()
+		return nil, err
+	}
+
 	// Initialize the stats fetcher that autopilot will use.
 	s.statsFetcher = NewStatsFetcher(logger, s.connPool, s.config.Datacenter)
 
-	// Initialize the authoritative ACL cache.
 	s.sentinel = sentinel.New(logger)
-	s.aclAuthCache, err = acl.NewCache(aclCacheSize, s.aclLocalFault, s.sentinel)
-	if err != nil {
-		s.Shutdown()
-		return nil, fmt.Errorf("Failed to create authoritative ACL cache: %v", err)
+	s.useNewACLs = 0
+	aclConfig := ACLResolverConfig{
+		Config:      config,
+		Delegate:    s,
+		CacheConfig: serverACLCacheConfig,
+		AutoDisable: false,
+		Logger:      logger,
+		Sentinel:    s.sentinel,
 	}
-
-	// Set up the non-authoritative ACL cache. A nil local function is given
-	// if ACL replication isn't enabled.
-	var local acl.FaultFunc
-	if s.IsACLReplicationEnabled() {
-		local = s.aclLocalFault
-	}
-	if s.aclCache, err = newACLCache(config, logger, s.RPC, local, s.sentinel); err != nil {
+	// Initialize the ACL resolver.
+	if s.acls, err = NewACLResolver(&aclConfig); err != nil {
 		s.Shutdown()
-		return nil, fmt.Errorf("Failed to create non-authoritative ACL cache: %v", err)
+		return nil, fmt.Errorf("Failed to create ACL resolver: %v", err)
 	}
 
 	// Initialize the RPC layer.
-	if err := s.setupRPC(tlsWrap); err != nil {
+	if err := s.setupRPC(tlsConfigurator.OutgoingRPCWrapper()); err != nil {
 		s.Shutdown()
 		return nil, fmt.Errorf("Failed to start RPC layer: %v", err)
 	}
@@ -402,15 +449,20 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		go s.Flood(nil, portFn, s.serfWAN)
 	}
 
+	// Start enterprise specific functionality
+	if err := s.startEnterprise(); err != nil {
+		s.Shutdown()
+		return nil, err
+	}
+
+	// Initialize Autopilot. This must happen before starting leadership monitoring
+	// as establishing leadership could attempt to use autopilot and cause a panic.
+	s.initAutopilot(config)
+
 	// Start monitoring leadership. This must happen after Serf is set up
 	// since it can fire events when leadership is obtained.
 	go s.monitorLeadership()
 
-	// Start ACL replication.
-	if s.IsACLReplicationEnabled() {
-		go s.runACLReplication()
-	}
-
 	// Start listening for RPC requests.
 	go s.listen(s.Listener)
 
@@ -422,9 +474,6 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 	// Start the metrics handlers.
 	go s.sessionStats()
 
-	// Initialize Autopilot
-	s.initAutopilot(config)
-
 	return s, nil
 }
 
@@ -457,6 +506,7 @@ func (s *Server) setupRaft() error {
 		MaxPool:               3,
 		Timeout:               10 * time.Second,
 		ServerAddressProvider: serverAddressProvider,
+		Logger:                s.logger,
 	}
 
 	trans := raft.NewNetworkTransportWithConfig(transConfig)
@@ -1016,9 +1066,31 @@ func (s *Server) Stats() map[string]map[string]string {
 		"serf_lan": s.serfLAN.Stats(),
 		"runtime":  runtimeStats(),
 	}
+
+	if s.ACLsEnabled() {
+		if s.UseLegacyACLs() {
+			stats["consul"]["acl"] = "legacy"
+		} else {
+			stats["consul"]["acl"] = "enabled"
+		}
+	} else {
+		stats["consul"]["acl"] = "disabled"
+	}
+
 	if s.serfWAN != nil {
 		stats["serf_wan"] = s.serfWAN.Stats()
 	}
+
+	for outerKey, outerValue := range s.enterpriseStats() {
+		if _, ok := stats[outerKey]; ok {
+			for innerKey, innerValue := range outerValue {
+				stats[outerKey][innerKey] = innerValue
+			}
+		} else {
+			stats[outerKey] = outerValue
+		}
+	}
+
 	return stats
 }
 
@@ -1040,6 +1112,12 @@ func (s *Server) GetLANCoordinate() (lib.CoordinateSet, error) {
 	return cs, nil
 }
 
+// ReloadConfig is used to have the Server do an online reload of
+// relevant configuration information
+func (s *Server) ReloadConfig(config *Config) error {
+	return nil
+}
+
 // Atomically sets a readiness state flag when leadership is obtained, to indicate that server is past its barrier write
 func (s *Server) setConsistentReadReady() {
 	atomic.StoreInt32(&s.readyForConsistentReads, 1)
